/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001,2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Original Author: Guruprasad Krishnamurthy <kgprasad@hotmail.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: request.c,v 1.39 2003/02/28 17:16:18 kds Exp $

   Contains the SDP service discovery client request handling code.

   Fixes:
   		Dmitry Kasatkin		- Continuation mechanism, MTU support
*/

#include <affix/config.h>

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <affix/sdp.h>
#include <affix/sdpclt.h>
#include <affix/sdpsrv.h>
#include "request.h"
#include "servicedb.h"
#include "utils.h"
#include "attr.h"
#include "cstate.h"

extern void updateSDPServerDBTimestamp(void);
extern slist_t	*serviceList;

/*
 ** Search/Attribute element sequence extractor. Builds
 ** a slist_t whose elements are those found in the 
 ** sequence. The data type of elements found in the
 ** sequence (now populated in slist_t) is returned
 ** in the reference pDataType
 */
int extractDataElementSequence(
		char *pduBuffer,
		slist_t **svcReqSeq,
		uint8_t *pDataType)
{
	int	status, bytesScanned;
	char	*pdata;
	uint8_t	seqType;
	int 	length = 0;
	short	numberOfElements;
	int	seqLength;
	slist_t	*pSeq;
	uint8_t	dataType;

	bytesScanned = 0;
	pdata = extractSequenceType(pduBuffer, &seqType, &length, &bytesScanned);
	DBPRT("Seq type : %d", seqType);
	if (!pdata || (SDP_DTD_TYPE(seqType) != SDP_DTD_SEQ)) {
		BTERROR("Unknown seq type");
		return -1;
	}
	DBPRT("Data size : %d", length);
	numberOfElements = 0;
	seqLength = 0; 
	pSeq = NULL;
	for (;;) {
		char	*pElem = NULL;
		int	localSeqLength = 0;

		dataType = __get_u8(pdata);
		DBPRT("Data type : %d", dataType);
		switch (dataType) {
			case UnsignedInteger_16 :
				pdata += sizeof(uint8_t);
				seqLength += sizeof(uint8_t);
				pElem = (char *)malloc(sizeof(uint16_t));
				if (!pElem)
					break;
				*(uint16_t *)pElem = ntohs(__get_u16(pdata));
				pdata += sizeof(uint16_t);
				seqLength += sizeof(uint16_t);
				break;
			case UnsignedInteger_32 :
				pdata += sizeof(uint8_t);
				seqLength += sizeof(uint8_t);
				pElem = (char *)malloc(sizeof(uint32_t));
				if (!pElem)
					break;
				*(uint32_t *)pElem = ntohl(__get_u32(pdata));
				pdata += sizeof(uint32_t);
				seqLength += sizeof(uint32_t);
				break;
			case UUID_16 :
			case UUID_32 :
			case UUID_128 :
				pElem = (char *)malloc(sizeof(uuid_t));
				if (!pElem)
					break;
				status = extractUUID(pdata, (uuid_t *)pElem, &localSeqLength);
				if (status) {
					free(pElem);
					return status;
				}
				seqLength += localSeqLength;
				pdata += localSeqLength;
				break;
		}
		if (!pElem)
			continue;
		s_list_append(&pSeq, pElem);
		numberOfElements++;
		DBPRT("No of elements : %d", numberOfElements);
		if (seqLength == length)
			break;
		else if (seqLength > length)
			return -1;
	}
	*svcReqSeq = pSeq;
	bytesScanned += seqLength;
	*pDataType = dataType;
	return bytesScanned;
}



/*
 ** Service search request PDU. This methods extracts the "search pattern"
 ** which is a seqquence of UUIDs calls the matching function
 ** to find matching services
 */
int handleServiceSearchRequest(sdp_request_t *requestPdu, sdppdu_t *pdu)
{
	int		status = 0, i;
	char		*pdata;
	int		bytesScanned = 0;
	slist_t		*searchPattern = NULL;
	int		maxExpectedServiceRecordCount, realServiceRecordCount;
	uint8_t		dataType;
	sdp_cstate_t	*pCState = NULL;
	char		*pCacheBuffer = NULL;
	int		handleSize = 0;
	long		cStateId = -1;
	short		responseCount = 0;
	short		*pTotalRecordCount;
	short		*pCurrentRecordCount;
	int		MTU;

	pdata = requestPdu->data + sizeof(sdp_hdr_t);
	bytesScanned = extractDataElementSequence(pdata, &searchPattern, &dataType);

	pdata += bytesScanned;
	maxExpectedServiceRecordCount = ntohs(__get_u16(pdata));
	DBPRT("Expected count: %d", maxExpectedServiceRecordCount);
	DBPRT("Bytes scanned : %d", bytesScanned);
	pdata += sizeof(uint16_t);

	/*
	 ** Check if continuation state exists, if yes attempt
	 ** to get response remainder from cache, else send error
	 */
	pCState = getContinuationState(pdata);

	MTU = requestPdu->mtu - sizeof(sdp_hdr_t) - sizeof(uint16_t) - sizeof(uint16_t) - sizeof(sdp_cstate_t);
	realServiceRecordCount = btmin(maxExpectedServiceRecordCount, (MTU >> 2));
	DBPRT("MTU: %d, realrec: %d", MTU, realServiceRecordCount);
	/*
	 ** Make space in the response buffer for total
	 ** and current record counts
	 */
	pdata = pdu->data;

	/*
	 ** Total service record count = 0
	 */
	pTotalRecordCount = (short *)pdata;
	__put_u16(pdata, 0);
	pdata += sizeof(uint16_t);
	pdu->length += sizeof(uint16_t);

	/*
	 ** current service record count = 0
	 */
	pCurrentRecordCount = (short *)pdata;
	__put_u16(pdata, 0);
	pdata += sizeof(uint16_t);
	pdu->length += sizeof(uint16_t);

	if (pCState == NULL) {
		/*
		 ** For every element in the sdpsvc_t DB,
		 ** do a pattern search
		 */
		int index = 0;
		handleSize = 0;
		for (;;) {
			void * data;
			sdpsvc_t *svcRec;

			data = s_list_nth_data(serviceList, index);
			if (data != NULL) {
				svcRec = (sdpsvc_t *)data;
				DBPRT("Checking svcRec : 0x%x", svcRec->serviceRecordHandle);
				if (sdp_match_uuid(searchPattern, svcRec->targetPattern)) {
					responseCount++; 
					__put_u32(pdata, htonl(svcRec->serviceRecordHandle));
					pdata += sizeof(uint32_t);
					handleSize += sizeof(uint32_t);
				}
				index++;
			} else
				break;
		}
		DBPRT("Match count : %d", responseCount);
		/*
		 ** Add "handleSize" to pdu buffer size
		 */
		pdu->length += handleSize;
		__put_u16(pTotalRecordCount,  htons(responseCount));
		__put_u16(pCurrentRecordCount, htons(responseCount));

		if (responseCount > realServiceRecordCount) {
			/* 
			 ** We need to cache the response and generate a 
			 ** continuation state 
			 */
			cStateId = createContinuationStateBuffer(requestPdu->fd, pdu);
			/*
			 ** Subtract the handleSize since we now send only
			 ** a subset of handles
			 */
			pdu->length -= handleSize;
		} else {
			/*
			 ** NULL continuation state
			 */
			setContinuationStatePDU(pdu, NULL);
		}
	}

	/*
	 ** Under both the conditions below, the response buffer is not
	 ** built up yet
	 */
	if ((pCState != NULL) || (cStateId != -1)) {
		short lastIndex = 0;

		if (pCState != NULL) {
			sdppdu_t *pCache;
			/*
			 ** Get the previous sdp_cstate_t and obtain
			 ** the cached response
			 */
			pCache = getCachedResponse(requestPdu->fd, pCState);
			if (!pCache) {
				status = -1;
				goto exit;
			}
			pCacheBuffer = pCache->data;
			responseCount = ntohs(__get_u16(pCacheBuffer));
			/*
			 ** Get the index of the last sdpsvc_t sent
			 */
			lastIndex = pCState->value.lastIndexSent;
		} else {
			pCacheBuffer = pdu->data;
			lastIndex = 0;
		}
		/*
		 ** Set the local buffer pointer to beyond the
		 ** current record count. And increment the cached
		 ** buffer pointer to beyond the counters ..
		 */
		pdata = (char *)pCurrentRecordCount + sizeof(uint16_t);

		/*
		 ** There is the toatalCount and the currentCount 
		 ** fields to increment beyond
		 */
		pCacheBuffer += (2 * sizeof(uint16_t));

		if (pCState == NULL) {
			handleSize = realServiceRecordCount << 2; //sizeof(uint32_t); 
			i = realServiceRecordCount;
		} else {
			handleSize = 0;
			for (i = lastIndex; (i-lastIndex) < realServiceRecordCount &&
					i < responseCount; i++) {
				__put_u32(pdata, __get_u32(((uint32_t*)pCacheBuffer) + i));
				pdata += sizeof(uint32_t);
				handleSize += sizeof(uint32_t);
			}
		}

		/*
		 ** Add "handleSize" to pdu buffer size
		 */
		pdu->length += handleSize;
		__put_u16(pTotalRecordCount,  htons(responseCount));
		__put_u16(pCurrentRecordCount, htons(i-lastIndex));

		if (i == responseCount) {
			/*
			 ** Set "null" continuationState & and remove it from cache
			 */
			setContinuationStatePDU(pdu, NULL);
			removeCachedResponse(requestPdu->fd, pCState);
		} else {
			/*
			 ** There is more to it .. set the lastIndexSent to
			 ** the new value and move on ..
			 */
			sdp_cstate_t newState;

			DBPRT("Setting non-NULL sdp_cs_t");
			memset(&newState, 0, sizeof(sdp_cstate_t));
			if (pCState != NULL) {
				memcpy(&newState, pCState, sizeof(sdp_cstate_t));
				newState.value.lastIndexSent = i;
			} else {
				newState.timestamp = cStateId;
				newState.value.lastIndexSent = i;
			}
			setContinuationStatePDU(pdu, &newState);
		}
	}
exit:
	if (searchPattern)
		s_list_destroy(&searchPattern);
	return status;
}


/*
 ** This function generates the PDU form of a service record for
 ** a specified set of attributes and not the entire set. This is
 ** driven by a service discovery client which asks for attributes
 ** of a service record by speicifying a subset of attribute 
 ** identifiers
 */
void generateServiceRecordPDUFormById(sdpsvc_t *svcRec, uint16_t attrId,
					sdppdu_t *attrSubsetPDUForm)
{
	sdpdata_t *pData;

	pData = (sdpdata_t*)sdp_get_attr(svcRec, attrId);
	if (!pData || !pData->pdu.data || !pData->pdu.length)
		return;
	appendToPDUForm(attrSubsetPDUForm, &pData->pdu);
}


/*
 ** Extract attribute identifiers from the request PDU.
 ** Clients could request a subset of attributes (by id)
 ** from a service record, instead of the whole set. The
 ** requested identifiers are present in the PDU form of
 ** the request
 */
int extractAttributesFromServiceRecord(
		sdpsvc_t *svcRec,
		slist_t *attrSeq,
		uint8_t attrDataType,
		sdppdu_t *pdu,
		int maxExpectedAttrSize
		)
{
	int 	status = 0;
	int 	attrLength;
	int 	index = 0;

	if (!svcRec)
		return SDP_ERR_SERVICE_RECORD_HANDLE;

	if (!attrSeq ) {
		DBPRT("Attribute sequence is NULL");
		return 0;
	}
#ifdef CONFIG_AFFIX_DEBUG
	if (attrSeq != NULL) {
		DBPRT("Entries in attr seq : %d", s_list_length(attrSeq));
	}
	DBPRT("AttrDataType : %d", attrDataType);
#endif

	/*
	 ** Set the maximum expected size before hand
	 */
	pdu->size = maxExpectedAttrSize;

	/*
	 ** This is a request for the entire attribute range
	 */
	if (svcRec->pdu.data == NULL) {
		DBPRT("Generating svc rec pdu form");
		generateServiceRecordPDUForm(svcRec);
	}

	attrLength = 0;

	if (attrDataType == UnsignedInteger_16) {
		void	*pAttr;
		for (index = 0; (pAttr = s_list_nth_data(attrSeq, index)); index++) {
			uint16_t	attrId = 0;

			attrId = *(uint16_t *)pAttr;
			DBPRT("Attr id requested : 0x%x", attrId);
			generateServiceRecordPDUFormById(svcRec, attrId, pdu);
		}
	} else if (attrDataType == UnsignedInteger_32) {
			uint8_t	completeAttrSetBuilt = 0;
			void	*pAttr;
			for (index = 0; (pAttr = s_list_nth_data(attrSeq, index)); index++) {
				uint32_t 	attrRange = 0;
				uint16_t 	lowIdLim = 0;
				uint16_t 	highIdLim = 0;

				attrRange = *(uint32_t *)pAttr;
				lowIdLim = ((0xffff0000 & attrRange) >> 16);
				highIdLim = 0x0000ffff & attrRange;

				DBPRT("attr range : 0x%x", attrRange);
				DBPRT("Low id : 0x%x", lowIdLim);
				DBPRT("High id : 0x%x", highIdLim);
				if ((lowIdLim == 0x0000) &&
						(highIdLim == 0xffff) &&
						(completeAttrSetBuilt == 0) &&
						(svcRec->pdu.length <= pdu->size)) {
					completeAttrSetBuilt = 1;
					/*
					 ** Create a copy of it
					 */
					DBPRT("Copy attr pdu form : %d", svcRec->pdu.length);
					memcpy(pdu->data, svcRec->pdu.data, 
							svcRec->pdu.length);
					pdu->length = svcRec->pdu.length;
				} else {
					/*
					 ** Not the entire range of attributes, pick and choose
					 ** attributes requested
					 */
					int attrId;
					for (attrId = lowIdLim; attrId <= highIdLim; attrId++)
						generateServiceRecordPDUFormById(svcRec, attrId, pdu);
				}
			}
	} else {
		status = SDP_ERR_REQUEST_SYNTAX;
		BTERROR("Unexpected data type : 0x%x", attrDataType);
		BTERROR("Expect uint16_t or uint32_t");
	}
	return status;
}


/*
 ** A request for the attributes of a service record.
 ** First check if the service record (specified by
 ** service record handle) exists, then call the attribute
 ** streaming function
 */
int handleServiceAttributeRequest(sdp_request_t *requestPdu, sdppdu_t *pdu)
{
	char		*pdata;
	int		status = 0;
	sdpsvc_t	*svcRec;
	int		maxResponseSize;
	uint32_t	svcRecHandle;
	sdp_cstate_t	*pCState;
	char		*pResponse = NULL;
	short 		continuationStateSize = 0;
	slist_t		*attrSeq = NULL;
	uint8_t		attrDataType = 0;
	int		bytesScanned = 0;

	pdata = requestPdu->data + sizeof(sdp_hdr_t);
	svcRecHandle = ntohl(__get_u32(pdata));
	pdata += sizeof(uint32_t);
	maxResponseSize = ntohs(__get_u16(pdata));
	pdata += sizeof(uint16_t);

	/*
	 ** Extract the attribute list
	 */
	bytesScanned = extractDataElementSequence(pdata, &attrSeq, &attrDataType);
	pdata += bytesScanned;

	/*
	 ** Check if continuation state exists, if yes attempt
	 ** to get response remainder from cache, else send error
	 */
	pCState = getContinuationState(pdata);

	DBPRT("SvcRecHandle : 0x%x", svcRecHandle);
	DBPRT("maxResponseSize : %d", maxResponseSize);

	maxResponseSize = btmin(maxResponseSize,
			(requestPdu->mtu-sizeof(sdp_hdr_t)-sizeof(uint16_t)-sizeof(sdp_cstate_t)));
	/*
	 * pull header for AttributeList byte count
	 */
	pdu->data += sizeof(uint16_t);
	pdu->size -= sizeof(uint16_t);

	if (pCState) {
		sdppdu_t *pCache = NULL;
		short bytesSent = 0;

		pCache = getCachedResponse(requestPdu->fd, pCState);
		DBPRT("Obtained cached response : %p", pCache);
		if (pCache != NULL) {
			pResponse = pCache->data;
			bytesSent = btmin(maxResponseSize, 
					(pCache->length - pCState->value.maxBytesSent));
			memcpy(pdu->data, (pResponse + pCState->value.maxBytesSent), 
					bytesSent);
			pdu->length += bytesSent;
			pCState->value.maxBytesSent += bytesSent;
			DBPRT("Response size : %d sending now : %d bytes sent so far : %d", 
					pCache->length, bytesSent, pCState->value.maxBytesSent);
			if (pCState->value.maxBytesSent == pCache->length) {
				/* remove from cache */
				continuationStateSize = setContinuationStatePDU(pdu, NULL);
				removeCachedResponse(requestPdu->fd, pCState);
			} else
				continuationStateSize = setContinuationStatePDU(pdu, pCState);
		} else {
			status = -1;
			BTERROR("NULL cache buffer and non-NULL continuation state");
		}
	} else {
		svcRec = findServiceRecord(svcRecHandle);
		//SDPXXX: is pdata correct ?????
		status = extractAttributesFromServiceRecord(svcRec, attrSeq, attrDataType, 
								pdu, pdu->size);
		//BTINFO("length: %d, max: %d\n", pdu->length, maxResponseSize);
		if (pdu->length > maxResponseSize) {
			sdp_cstate_t newState;

			memset((char *)&newState, 0, sizeof(sdp_cstate_t));
			newState.timestamp = createContinuationStateBuffer(requestPdu->fd, pdu);
			/*
			 ** Reset the buffer size to the maximum expected and
			 ** set the sdp_cstate_t
			 */
			DBPRT("Creating continuation state of size : %d", pdu->length);
			pdu->length = maxResponseSize;
			newState.value.maxBytesSent = maxResponseSize;
			continuationStateSize = setContinuationStatePDU(pdu, &newState);
		} else
			continuationStateSize = setContinuationStatePDU(pdu, NULL);
	}
	// push header
	pdu->data -= sizeof(uint16_t);
	pdu->size += sizeof(uint16_t);
	if (status != 0) {
		__put_u16(pdu->data, htons(status));
		pdu->length = sizeof(uint16_t);
	} else {
		/*
		 * set attribute list byte count
		 */
		__put_u16(pdu->data, htons(pdu->length - continuationStateSize));
		pdu->length += sizeof(uint16_t);
	}
	if (attrSeq)
		s_list_destroy(&attrSeq);
	return status;
}


/*
 ** The single request for a combined "service search" and
 ** "attribute extraction"
 */
int handleServiceSearchAttributeRequest(sdp_request_t *requestPdu, sdppdu_t *pdu)
{
	int		status = 0;
	char		*pdata;
	int		bytesScanned = 0;
	slist_t		*searchPattern = NULL;
	int		maxExpectedAttrSize;
	uint8_t		dataType;
	sdp_cstate_t	*pCState = NULL;
	char		*pResponse = NULL;
	short		continuationStateSize = 0;
	slist_t		*attrSeq = NULL;
	uint8_t		attrDataType = 0;
	int		index = 0;
	int		responseCount = 0;
	sdppdu_t	localPDUForm;

	DBPRT("In svc srch attr request");
	pdata = requestPdu->data + sizeof(sdp_hdr_t);

	bytesScanned = extractDataElementSequence(pdata, &searchPattern, &dataType);
	DBPRT("Bytes scanned : %d", bytesScanned);

	pdata += bytesScanned;
	maxExpectedAttrSize = ntohs(__get_u16(pdata));
	pdata += sizeof(uint16_t);
	DBPRT("Max Attr expected : %d", maxExpectedAttrSize);

	/*
	 ** Extract the attribute list
	 */
	bytesScanned = extractDataElementSequence(pdata, &attrSeq, &attrDataType);
	pdata += bytesScanned;

	/*
	 ** Check if continuation state exists, if yes attempt
	 ** to get response remainder from cache, else send error
	 */
	pCState = getContinuationState(pdata);	//this is my Continuation Information

	localPDUForm.data = malloc(USHRT_MAX);
	if (!localPDUForm.data)
		goto exit;
	localPDUForm.length = 0;
	localPDUForm.size = USHRT_MAX;

	maxExpectedAttrSize = btmin(maxExpectedAttrSize,
			(requestPdu->mtu-sizeof(sdp_hdr_t)-sizeof(sdp_cstate_t)-sizeof(uint16_t)));
	/*
	 * pull header for AttributeList byte count
	 */
	pdu->data += sizeof(uint16_t);
	pdu->size -= sizeof(uint16_t);

	if (pCState == NULL) {
		/*
		 * No Continuation State -> Create New Response
		 */
		for (;;) {
			void * data;
			sdpsvc_t *svcRec;

			localPDUForm.length = 0;
			memset(localPDUForm.data, 0, USHRT_MAX);

			data = s_list_nth_data(serviceList, index);
			if (!data)
				break;
			svcRec = (sdpsvc_t *)data;
			if (sdp_match_uuid(searchPattern, svcRec->targetPattern)) {
				responseCount++;
				status = extractAttributesFromServiceRecord(svcRec, attrSeq, attrDataType, 
									&localPDUForm, localPDUForm.size);
				DBPRT("Response count : %d", responseCount);
				DBPRT("Local PDU size : %d", localPDUForm.length);
				if (status == -1) {
					DBPRT("Extract attr from svc rec returns err");
					break;
				}
				if (pdu->length + localPDUForm.length < pdu->size) {
					// to be sure no relocations
					appendToPDUForm(pdu, &localPDUForm);
				} else {
					BTERROR("Relocation needed");
					break;
				}
				DBPRT("Net PDU size : %d", pdu->length);
			}
			index++;
		}
		if (pdu->length > maxExpectedAttrSize) {
			sdp_cstate_t newState;

			memset((char *)&newState, 0, sizeof(sdp_cstate_t));
			newState.timestamp = createContinuationStateBuffer(requestPdu->fd, pdu);
			/*
			 ** Reset the buffer size to the maximum expected and
			 ** set the sdp_cstate_t
			 */
			pdu->length = maxExpectedAttrSize;
			newState.value.maxBytesSent = maxExpectedAttrSize;
			continuationStateSize = setContinuationStatePDU(pdu, &newState);
		} else
			continuationStateSize = setContinuationStatePDU(pdu, NULL);
	} else {
		/*
		 * Continuation State exists -> get from cache
		 */
		sdppdu_t *pCache;
		uint16_t bytesSent = 0;

		/*
		 ** Get the cached response from the buffer
		 */
		pCache = getCachedResponse(requestPdu->fd, pCState);
		if (pCache != NULL) {
			pResponse = pCache->data;
			bytesSent = btmin(maxExpectedAttrSize, 
					(pCache->length - pCState->value.maxBytesSent));
			memcpy(pdu->data, (pResponse + pCState->value.maxBytesSent), 
					bytesSent);
			pdu->length += bytesSent;
			pCState->value.maxBytesSent += bytesSent;
			if (pCState->value.maxBytesSent == pCache->length) {
				/* remove response from cache as well */
				continuationStateSize = setContinuationStatePDU(pdu, NULL);
				removeCachedResponse(requestPdu->fd, pCState);
			} else
				continuationStateSize = setContinuationStatePDU(pdu, pCState);
		} else {
			status = -1;
			DBPRT("Non null continuation state, but null cache buffer");
		}
	}
	if ((responseCount == 0) && (pCState == NULL)) {
		//we've not found anything - create an empty sequence
		appendToPDUForm(pdu, &localPDUForm);
	}
	// push header
	pdu->data -= sizeof(uint16_t);
	pdu->size += sizeof(uint16_t);
	if (status != 0) {
		__put_u16(pdu->data, htons(status));
		pdu->length = sizeof(uint16_t);
	} else {
		/*
		 * set attribute list byte count
		 */
		__put_u16(pdu->data, htons(pdu->length - continuationStateSize));
		pdu->length += sizeof(uint16_t);
	}
	free(localPDUForm.data);
exit:
	if (searchPattern)
		s_list_destroy(&searchPattern);
	if (attrSeq)
		s_list_destroy(&attrSeq);
	return status;
}

/* 
  MAIN in service registraion
 */
sdpsvc_t *serverExtractPDU(char *pdata, uint32_t handleExpected, int *bytesScanned)
{
	int		localExtractedLength = 0;
	char		*localSeqPtr = NULL;
	uint8_t		dtd;
	int		sequenceLength = 0;
	sdpsvc_t	*svcRec = NULL;

	localSeqPtr = extractSequenceType(pdata, &dtd, &sequenceLength, bytesScanned);

	/*
	 ** Check if the PDU already has a service record handle,
	 ** int which case this may be a update request
	 */
	if (handleExpected != 0xffffffff)
		svcRec = findServiceRecord(handleExpected);
	
	if (svcRec == NULL) {
		svcRec = sdp_create_svc();
		svcRec->attributeList = NULL;
	}
	localExtractedLength = extractAttributes(svcRec, localSeqPtr, sequenceLength);
	if (localExtractedLength >= 0)
		*bytesScanned += localExtractedLength;
	return svcRec;
}

/*
 ** Request to create a brand new service record
 ** The newly created service record is added to the
 ** service repository
 * MAIN in service registraion
 */
int handleServiceRegistrationRequest(sdp_request_t *sdpReq, sdppdu_t *response)
{
	char		*pdata;
	int 		bytesScanned = 0;
	sdpsvc_t	*svcRec;
	sdp_hdr_t 	*sdpPduHeader;
	int 		length = 0;
	sdpdata_t 	*browseAttrData = NULL;

	/*
	 ** Extract buffer contents
	 */
	sdpPduHeader = (sdp_hdr_t *)sdpReq->data;
	pdata = sdpReq->data + sizeof(sdp_hdr_t);

	/*
	 ** Extract and save image of PDU, since we need
	 ** it when clients request thsi attribute
	 */
	svcRec = serverExtractPDU(pdata, 0xffffffff, &bytesScanned);
	if (!svcRec) {
		__put_u16(response->data, htons(SDP_ERR_REQUEST_SYNTAX));
		response->length = sizeof(uint16_t);
		return -1;
	}
	svcRec->serviceRecordHandle = (uint32_t)svcRec;
	addServiceRecord(svcRec);
	sdp_append_attr(svcRec, AttrServiceRecordHandle, sdp_put_u32(svcRec->serviceRecordHandle));
	/*
	 ** Also check if the browse group descriptor is NULL,
	 ** ensure that the record belongs to the ROOT group
	 */
	if ((browseAttrData = sdp_get_attr(svcRec, AttrBrowseGroupList)) == NULL)
		addUUID16ToTargetPattern(svcRec, PublicBrowseGroup);

	generateServiceRecordPDUForm(svcRec);
	updateSDPServerDBTimestamp();

	/*
	 ** Build a response buffer
	 */
	pdata = response->data;
	__put_u32(pdata, htonl(svcRec->serviceRecordHandle));
	DBPRT("Sending back svcRecHandle : 0x%x", svcRec->serviceRecordHandle);
	length += sizeof(uint32_t);
	response->length = length;
	
	/* set owner of the service */
	svcRec->fd = sdpReq->fd;
	return 0;
}


/*
 ** Register a service record (this means updating the
 ** contents of a newly created service record or updating
 ** the contents of an existing record)
 */
int handleServiceUpdateRequest(sdp_request_t *sdpReq, sdppdu_t *response)
{
	char *pdata;
	int status = 0;
	uint32_t svcRecHandle;
	sdpsvc_t *svcRecNew = NULL;
	sdpsvc_t *svcRecOld = NULL;
	sdp_hdr_t *sdpPduHeader;
	int bytesScanned = 0;

	/*
	 ** Extract service record handle
	 */
	sdpPduHeader = (sdp_hdr_t *)sdpReq->data;
	pdata = sdpReq->data + sizeof(sdp_hdr_t);
	svcRecHandle = ntohl(__get_u32(pdata));
	DBPRT("Svc Rec Handle : 0x%x\n", svcRecHandle);
	pdata += sizeof(uint32_t);

	svcRecOld = findServiceRecord(svcRecHandle);
	DBPRT("SvcRecOld : 0x%x\n", (uint32_t)svcRecOld);
	if (svcRecOld != NULL) {
		svcRecNew = serverExtractPDU(pdata, svcRecHandle, &bytesScanned);
		if ((svcRecNew != NULL) && 
				(svcRecHandle == svcRecNew->serviceRecordHandle)) {
			generateServiceRecordPDUForm(svcRecNew);
			updateSDPServerDBTimestamp();
		} else {
			DBPRT("SvcRecHandle : 0x%x\n", svcRecHandle);
			DBPRT("SvcRecHandleNew : 0x%x\n", svcRecNew->serviceRecordHandle);
			DBPRT("SvcRecNew : 0x%x\n", (uint32_t)svcRecNew);
			DBPRT("SvcRecOld : 0x%x\n", (uint32_t)svcRecOld);
			DBPRT("Failure to update, restore old value\n");
			if (svcRecNew != NULL)
				sdp_free_svc(svcRecNew);
			status = SDP_ERR_REQUEST_SYNTAX;
		}
	}
	else
		status = SDP_ERR_SERVICE_RECORD_HANDLE;

	pdata = response->data;
	__put_u16(pdata, htons(status));
	response->length = sizeof(uint16_t);
	return status;
}


/*
 ** Remove a registered service record
 */
int handleServiceRemovalRequest(sdp_request_t *sdpReq, sdppdu_t *response)
{
	char *pdata;
	int status = 0;
	uint32_t svcRecHandle;
	sdpsvc_t *svcRec;

	/*
	 ** Extract service record handle
	 */
	pdata = sdpReq->data + sizeof(sdp_hdr_t);
	svcRecHandle = ntohl(__get_u32(pdata));
	pdata += sizeof(uint32_t);

	svcRec = findServiceRecord(svcRecHandle);
	if (svcRec != NULL) {
		status = removeServiceRecord(svcRecHandle);
		sdp_free_svc(svcRec);
		if (status == 0)
			updateSDPServerDBTimestamp();
	}
	else {
		status = SDP_ERR_SERVICE_RECORD_HANDLE;
		DBPRT("Could not find record : 0x%x\n", svcRecHandle);
	}

	pdata = response->data;
	__put_u16(pdata, htons(status));
	response->length = sizeof(uint16_t);

	return status;
}


/*
 ** Top level request processor. Calls the appropriate processing
 ** function based on request type. Handles service registration
 ** client requests also
 */
void processRequest(sdp_request_t *request)
{
	char		*pdata;
	int		status = 0;
	sdp_hdr_t	*requestPduHeader;
	sdp_hdr_t	*responsePduHeader;
	sdppdu_t	responsePdu;
	char		*bufferStart;
	int		bytesSent = 0;

	DBPRT("In process req");
	requestPduHeader = (sdp_hdr_t *)request->data;

	bufferStart = malloc(USHRT_MAX);
	if (!bufferStart) {
		// FIXME: send response back
		return;
	}
	memset(bufferStart, 0, USHRT_MAX);
	responsePdu.data = bufferStart + sizeof(sdp_hdr_t);
	responsePdu.length = 0;
	responsePdu.size = USHRT_MAX - sizeof(sdp_hdr_t);
	responsePduHeader = (sdp_hdr_t *)bufferStart;

	switch (requestPduHeader->pduId) {
		case SDP_ServiceSearchRequest :
			DBPRT("Got a svc srch req");
			status = handleServiceSearchRequest(request, &responsePdu);

			responsePduHeader->pduId =
				((status == 0) ? SDP_ServiceSearchResponse : SDP_ErrorResponse);
			break;

		case SDP_ServiceAttributeRequest :
			DBPRT("Got a svc attr req");
			status = handleServiceAttributeRequest(request, &responsePdu);

			responsePduHeader->pduId =
				((status == 0) ? SDP_ServiceAttributeResponse : SDP_ErrorResponse);
			break;

		case SDP_ServiceSearchAttributeRequest :
			DBPRT("Got a svc srch attr req");
			status = handleServiceSearchAttributeRequest(request, &responsePdu);

			responsePduHeader->pduId = ((status == 0) ? 
					SDP_ServiceSearchAttributeResponse : SDP_ErrorResponse);
			break;

		/* Service Registration */

		case SDP_ServiceRegistrationRequest :
			DBPRT("Service register request");
			status = handleServiceRegistrationRequest(request, &responsePdu);

			responsePduHeader->pduId = ((status == 0) ? 
					SDP_ServiceRegistrationResponse : SDP_ErrorResponse);
			break;

		case SDP_ServiceUpdateRequest :
			DBPRT("Service update request");
			status = handleServiceUpdateRequest(request, &responsePdu);

			responsePduHeader->pduId = ((status == 0) ? 
					SDP_ServiceUpdateResponse : SDP_ErrorResponse);
			break;

		case SDP_ServiceRemovalRequest :
			DBPRT("Service removal request");
			status = handleServiceRemovalRequest(request, &responsePdu);

			responsePduHeader->pduId = ((status == 0) ? 
					SDP_ServiceRemovalResponse : SDP_ErrorResponse);
			break;

		default :
			BTERROR("Unknown PDU ID : 0x%x received", requestPduHeader->pduId);
			pdata = responsePdu.data;
			__put_u16(pdata, htons(SDP_ERR_REQUEST_SYNTAX));
			responsePdu.length = sizeof(uint16_t);
			responsePduHeader->pduId = SDP_ErrorResponse;
			break;
	}

	DBPRT("Status : %d", status);
	DBPRT("Sending a response back");
	responsePduHeader->transactionId = requestPduHeader->transactionId;
	responsePduHeader->paramLength = htons(responsePdu.length);
	/*
	 ** Now point back to the real buffer start and set the
	 ** real response length
	 */
	responsePdu.length += sizeof(sdp_hdr_t);
	responsePdu.data = bufferStart;

	/*
	 ** Stream the response PDU
	 */
	bytesSent = send(request->fd, responsePdu.data, responsePdu.length, 0);
	DBPRT("Bytes Sent : %d, requested: %d", bytesSent, responsePdu.length);
	free(bufferStart);
}

/*
 * Remove services by RequestSocket
 *
 */
int handleServiceDisconnectRequest(int fd)
{
	int	i, status;

	removeCachedResponse(fd, NULL);
start_again:
	for (i = 0; i < s_list_length(serviceList); i++) {
		sdpsvc_t *svcRec;

		svcRec = (sdpsvc_t *)s_list_nth_data(serviceList, i);
		DBPRT("fd: %d, svc->fd: %d\n", fd, svcRec->fd);
		if (svcRec->fd == fd) {
			status = removeServiceRecord(svcRec->serviceRecordHandle);
			sdp_free_svc(svcRec);
			goto start_again;
		}
	}
	return 0;	
}

